"""
===========================================
Collections (:mod:`data_tools.collections`)
===========================================

Time Series Data
===========
.. autosummary::
   :toctree: generated/

   TimeSeries    -- Enhanced `ndarray` with powerful data analysis features

"""
from .time_series import TimeSeries


__all__ = [
    "TimeSeries",
]
